/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2000 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 *
 * This video driver is STRONGLY derived from the video drivers of the
 * VSyncMAME project by S.Sakamaki (Saka)
 */

#include "compil.h"
#include "laguna.h"
#include "card.h"

static BYTE LagunaMMIO_inb(DWORD d0);
static void LagunaMMIO_outb(DWORD d0, BYTE d1);
static WORD LagunaMMIO_inw(DWORD d0);
static void LagunaMMIO_outw(DWORD d0, WORD d1);
#ifndef __PCIREAL__
static void LagunaLFBWritel(DWORD d0, DWORD d1);
#endif
static void ClearLagunaHorzAdjustReg(void);
static DWORD GetLagunaBytePerPixel(void);
static void SetLagunaFetchesPerLine(DWORD d0);
#ifndef __PCIREAL__
static void LagunaZeroMemory(void);
#endif
static void SetLagunaExtRegister(card_crtc STACK_PTR *cp);
static void SetLagunaInterlace(int flag, card_crtc STACK_PTR *cp);
static void SetLagunaVClock(int Num, int DeN, int PS);
static long GetLagunaVclkParam(long vclk, int STACK_PTR *Num, int STACK_PTR *DeN, int STACK_PTR *PS);
static void SetLagunaDoubleScan(int flag);
static void SetLagunaPLL(void);

static unsigned LagunaBusDeviceFunc;
static unsigned long LagunaLFBPhysicalAddress = 0;
#ifndef __PCIREAL__
static unsigned long LagunaLFBLinearAddress = 0;
static unsigned LagunaLFBLinearSelector = 0;
#endif

const char* driverLaguna(void) {
	return "Cirrus Logic GD5462/GD5464/GD5465 Laguna";
}

int detectLaguna(void)
{
	DWORD reg;

	if (pci_detect()!=0) {
		CARD_LOG(("laguna: PCI BIOS not Installed.\n"));
		return 0;
	}

	while(1){
		if (pci_find_device(0x00001013,0x000000D0,&LagunaBusDeviceFunc)==0) {
			CARD_LOG(("laguna: CL-GD5462 Found\n"));
			break;
		}
		if (pci_find_device(0x00001013,0x000000D4,&LagunaBusDeviceFunc)==0) {
			CARD_LOG(("laguna: CL-GD5464 Found\n"));
			break;
		}
		if (pci_find_device(0x00001013,0x000000D6,&LagunaBusDeviceFunc)==0) {
			CARD_LOG(("laguna: CL-GD5465 Found\n"));
			break;
		}

		CARD_LOG(("laguna: not found\n"));
		return 0;
	}

	if (pci_MMIO_address_map(LagunaBusDeviceFunc,0x14,0xffff8000)!=0) {
		CARD_LOG(( "laguna: pci_MMIO_address_map error\n"));
		return 0;
	}

	if (pci_read_dword(LagunaBusDeviceFunc,0x10,&reg)!=0) {
		CARD_LOG(( "laguna: pci_read_dword\n"));
		return 0;
	}

	LagunaLFBPhysicalAddress = reg & 0xFE000000;
	CARD_LOG(("laguna: LFB Address %08lx\n", LagunaLFBPhysicalAddress));

#ifndef __PCIREAL__
	if (card_create_linear_mapping(&LagunaLFBLinearAddress, LagunaLFBPhysicalAddress, 32 * 1024 * 1024) != 0){
		CARD_LOG(("laguna: card_create_linear_mapping error\n"));
		return 0;
	}
	CARD_LOG(("laguna: LFB Linear Address %08lx\n", LagunaLFBLinearAddress));

	if (card_create_selector(&LagunaLFBLinearSelector, LagunaLFBLinearAddress, 32 * 1024 * 1024) != 0) {
		card_remove_linear_mapping(&LagunaLFBLinearAddress);
		CARD_LOG(("laguna: card_create_selector error\n"));
		return 0;
	}
#endif

	return 1;
}

void resetLaguna(void)
{
	pci_MMIO_address_unmap();
#ifndef __PCIREAL__
	card_remove_linear_mapping(&LagunaLFBLinearAddress);
	card_remove_selector(&LagunaLFBLinearSelector);
#endif
}

int setLaguna(const card_crtc STACK_PTR* ccp)
{
	card_crtc cp = *ccp;
	int Num,DeN,PS;

	cp.dotclockHz = GetLagunaVclkParam(cp.dotclockHz, &Num, &DeN, &PS);

	cp.HBStart = cp.HSStart;
	cp.HBEnd   = cp.HTotal - 8;
	cp.VBStart = cp.VDisp;
	cp.VBEnd   = cp.VTotal - 1;

	if (pci_MMIO_address_relocate()!=0) {
		CARD_LOG(("laguna: pci_address_relocate error\n"));
		return 0;
	}

	card_signal_disable();
	card_generic_all_set(&cp);
	SetLagunaExtRegister(&cp);
	SetLagunaVClock(Num, DeN, PS);
	ClearLagunaHorzAdjustReg();
	SetLagunaFetchesPerLine(GetLagunaBytePerPixel() * cp.HDisp);
#ifndef __PCIREAL__
	LagunaZeroMemory();
#endif
	card_signal_enable();

	pci_MMIO_address_restore();

	return 1;
}

static void SetLagunaPLL(void)
{
	BYTE d0;

	d0 = card_in(0x3CC);
	d0 = d0 | 0x0C;
	card_out(0x3C2, d0);
}

static void SetLagunaExtRegister(card_crtc STACK_PTR *cp)
{
	int d0;

	d0 = card_crtc_get(0x1B);
	d0 = card_bitmov(d0, 5, 1, 0);
	d0 = card_bitmov(d0, 7, 1, 0);
	card_crtc_set(0x1B, d0);                                           // enable Blank End ext

	d0 = 0;
	d0 = card_bitmov(d0, 0, cp->VSStart                   , 10);
	d0 = card_bitmov(d0, 1, cp->VBStart                   , 10);
	d0 = card_bitmov(d0, 2, cp->VDisp                     , 10);
	d0 = card_bitmov(d0, 3, cp->VTotal                    , 10);
	d0 = card_bitmov(d0, 4, (cp->HSStart >> 3)    , 8);
	d0 = card_bitmov(d0, 5, (cp->HBStart >> 3)    , 8);
	d0 = card_bitmov(d0, 6, (cp->HDisp   >> 3) - 1, 8);
	d0 = card_bitmov(d0, 7, (cp->HTotal  >> 3) - 5, 8);
	card_crtc_set(0x1E, d0);                                           // ext horz vert reg

	d0 = card_crtc_get(0x1A);
	d0 = card_bitmov(d0, 4, (cp->HBEnd >> 3)      , 6);
	d0 = card_bitmov(d0, 5, (cp->HBEnd >> 3)      , 7);
	d0 = card_bitmov(d0, 6, cp->VBEnd                     , 8);
	d0 = card_bitmov(d0, 7, cp->VBEnd                     , 9);
	card_crtc_set(0x1A, d0);                                           // ext Blank End reg

	SetLagunaDoubleScan(cp->doublescan);
	SetLagunaInterlace(cp->interlace, cp);
	card_polarity_set(cp->hpolarity, cp->vpolarity);
}

static void SetLagunaDoubleScan(int flag)
{
	if(flag == 0){
		card_crtc_set(0x09, card_bitmov(card_crtc_get(0x09), 7, 0, 0));
	}else{
		card_crtc_set(0x09, card_bitmov(card_crtc_get(0x09), 7, 1, 0));
	}
}

static void SetLagunaInterlace(int flag, card_crtc STACK_PTR *cp)
{
	int d0;

	if(flag == 0){
		d0 = card_crtc_get(0x1A);
		if((d0 & 0x01) != 0){
			d0 = d0 & ~0x01;
			card_crtc_set(0x1A, d0);
		}
	}else{
		d0 = ((cp->HTotal >> 3) - 5) / 2;
		d0 = (d0 * cp->interlaceratio) / 100;
		card_crtc_set(0x19, d0);                                           // interlace offset
		d0 = card_crtc_get(0x1A);
		if((d0 & 0x01) == 0){
			d0 = d0 | 0x01;
			card_crtc_set(0x1A, d0);
		}
	}
}

static void ClearLagunaHorzAdjustReg(void)
{
	BYTE d0;

	d0 = LagunaMMIO_inb(0x70);
	d0 = d0 & 0xC0;
	LagunaMMIO_outb(0x70, d0);
}

static DWORD GetLagunaBytePerPixel(void)
{
	DWORD d0;

	d0 = (LagunaMMIO_inw(0xC0) >> 12) & 0x03;
	d0 += 1;
	return d0;
}

static void SetLagunaFetchesPerLine(DWORD d0)
{
	WORD d1;

	d0 = d0 + 127;
	d0 = d0 / 128;
	d0 = d0 & 0x3F;
	d1 = LagunaMMIO_inw(0xEA) & ~0x3F00;
	d1 = d1 | (d0 << 8);
	LagunaMMIO_outw(0xEA, d1);
}

#ifndef __PCIREAL__
static void LagunaZeroMemory(void)
{
	__dpmi_regs     r;
	DWORD installedmemory;
	unsigned loopcounter;

	r.h.ah = 0x12;
	r.h.bl = 0x85;
	__dpmi_int(0x10, &r);
	installedmemory = r.h.al * 65536;
	CARD_LOG(("laguna: Installed Memory %d MByte\n", installedmemory / (1024 * 1024) ));

	loopcounter = 0;
	while(loopcounter < installedmemory){
		LagunaLFBWritel(loopcounter, 0);
		loopcounter += 4;
	}
}
#endif

static long GetLagunaVclkParam(long vclk, int STACK_PTR *Num, int STACK_PTR *DeN, int STACK_PTR *PS)
{
	double desfreq;
	double temp;
	double tempdt;
	double dt;
	int d0;
	int d1;

	if(vclk > 175000000) vclk = 175000000;
	*PS     = 1;
	if(vclk >=      75000000){
		*PS  = 1;
	}else if(vclk >=  37500000){
		*PS  = 2;
	}else if(vclk >=  13500000){
		*PS  = 4;
	}else{
		*PS  = 8;
	}

	desfreq = vclk * (*PS);
	dt = 250000000.0;
	*Num = 0;
	*DeN = 0;
	d0 = 2;
	while(d0 < 126){
		d1 = 2;
		while(d1 < 126){
			temp   = 14318000 * (double)d0 / d1;
			tempdt = desfreq - temp;
			if(tempdt < 0) tempdt *= -1;
			if(tempdt < dt){
				*Num = d0;
				*DeN = d1;
				dt = tempdt;
			}
			d1 += 1;
		}
		d0 += 1;
	}
	CARD_LOG(("Num %3d, DeN %3d, PS %1d\n", *Num, *DeN, *PS));
	CARD_LOG(("14.318 * (Num/DeN) : %lf\n", 14.318 *((*Num)/(*DeN))));
	CARD_LOG(("14.318 * (Num/DeN)/PS : %lf\n", 14.318 * ((double)(*Num)/(*DeN))/(*PS)));
	vclk = 14318000 * ((double)(*Num)/(*DeN) / (*PS));
	CARD_LOG(("vclk : %lf\n", (double)vclk / 1000000.0));

	return vclk;
}

static void SetLagunaVClock(int Num, int DeN, int PS)
{
	DWORD _PS;
	WORD d0;
	BYTE _Num;
	BYTE _DeN;

	SetLagunaPLL();

	_PS = 0;
	while(1){
		PS = PS >> 1;
		if(PS == 0) break;
		_PS += 1;
	}

	_Num = Num & 0x7F;
	_DeN = DeN << 1;
	card_seq_set(0x1E, _Num);
	card_seq_set(0x0E, _DeN);

	d0 = LagunaMMIO_inw(0xC0) & ~0xC000;
	d0 = d0 | (_PS << 14);
	LagunaMMIO_outw(0xC0, d0);
}

static BYTE LagunaMMIO_inb(DWORD d0)
{
	return _farpeekb(pci_MMIO_selector_get(), d0);
}

static void LagunaMMIO_outb(DWORD d0, BYTE d1)
{
	_farpokeb(pci_MMIO_selector_get(), d0, d1);
}

static WORD LagunaMMIO_inw(DWORD d0)
{
	return _farpeekw(pci_MMIO_selector_get(), d0);
}

static void LagunaMMIO_outw(DWORD d0, WORD d1)
{
	_farpokew(pci_MMIO_selector_get(), d0, d1);
}

#ifndef __PCIREAL__
static void LagunaLFBWritel(DWORD d0, DWORD d1)
{
	_farpokel(LagunaLFBLinearSelector, d0, d1);
}
#endif
