/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2000 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 */

#include "screen.h"
#include "scrvga.h"
#include "scrvbe.h"
#include "vga.h"
#include "vgaline.h"
#include "vgatweak.h"
#include "vbe.h"
#include "vbeline.h"

#include "../card/card.h"

#include <stdarg.h>
#include <sys/farptr.h>
#include <dpmi.h>
#include <stdio.h>
#include <dos.h>
#include <assert.h>
#include <time.h>
#include <stdlib.h>

/* State */
video_internal video_state;
unsigned FAST (*video_write_selector)(void);
unsigned FAST (*video_write_line)(unsigned y);

static void pci_scan_device_callback(unsigned bus_device_func, unsigned vendor, unsigned device) {
	DWORD dw;
	unsigned base_class;
	unsigned subsys_card;
	unsigned subsys_vendor;

	if (pci_read_dword(bus_device_func,0x8,&dw)!=0)
		return;

	base_class = (dw >> 24) & 0xFF;
	if (base_class != 0x3 /* PCI_CLASS_DISPLAY */)
		return;

	if (pci_read_dword(bus_device_func,0x2c,&dw)!=0)
		return;

	subsys_vendor = dw & 0xFFFF;
	subsys_card = (dw >> 16) & 0xFFFF;

	video_log("video: found pci display vendor:%04x, device:%04x, subsys_vendor:%04x, subsys_card:%04x\n",vendor,device,subsys_vendor,subsys_card);
}

/* Inizialize the video system
	Return the list of video mode avaliable
*/
void video_init(void) {
	assert( !video_active() );

	video_log("video: video_init\n");

	if (pci_scan_device(pci_scan_device_callback)!=0) {
		video_log("video: error scanning pci display device, resume and continue\n");
	}

	vbe_init();
	vga_init();
	vbeline_init();

	memset(&video_state,0,sizeof(video_state));

	/* enable */
	video_state.active = 1;

	/* grab the actual video mode */
	if (video_mode_grab(&video_state.old_mode_info) == 0) {
		strcpy(video_state.old_mode_info.name,"startup_mode");
		video_state.old_mode_grabbed = 1;
	} else {
		video_state.old_mode_grabbed = 0;
	}
}

/* Deinizialize the video system */
void video_done() {
	assert( video_active() );

	video_log("video: video_done\n");

	if (video_mode_active())
		video_mode_done();
	if (vbeline_active())
		vbeline_done();
	if (vbe_active())
		vbe_done();
	if (vga_active())
		vga_done();

	/* disable */
	video_state.active = 0;
}

/***************************************************************************/
/* Options */

static int video_option_vbeline_tweak = VIDEO_OPTION_VBELINE_TWEAK_AGGRESSIVE;

void video_option_vbeline_tweak_set(int mode) {
	assert( mode == VIDEO_OPTION_VBELINE_TWEAK_MODERATE || mode == VIDEO_OPTION_VBELINE_TWEAK_AGGRESSIVE );
	video_option_vbeline_tweak = mode;
}

int video_option_vbeline_tweak_get(void) {
	return video_option_vbeline_tweak;
}

static int video_option_vbeline_driver = VIDEO_OPTION_VBELINE_DRIVER_AUTO;

void video_option_vbeline_driver_set(int mode) {
	video_option_vbeline_driver = mode;
}

int video_option_vbeline_driver_get(void) {
	return video_option_vbeline_driver;
}

static int video_option_default_mode = VIDEO_OPTION_DEFAULT_MODE_SMART;

void video_option_default_mode_set(int mode) {
	assert( mode == VIDEO_OPTION_DEFAULT_MODE_ALL || mode == VIDEO_OPTION_DEFAULT_MODE_SMART );
	video_option_default_mode = mode;
}

int video_option_default_mode_get(void) {
	return video_option_default_mode;
}

static int video_option_crtc_blank = VIDEO_OPTION_CRTC_BLANK_AGGRESSIVE;

void video_option_crtc_blank_set(int mode) {
	assert( mode == VIDEO_OPTION_CRTC_BLANK_AGGRESSIVE || mode == VIDEO_OPTION_CRTC_BLANK_MODERATE );
	video_option_crtc_blank = mode;
}

int video_option_crtc_blank_get(void) {
	return video_option_crtc_blank;
}

static double video_option_vbeline_clock_multiplier = 1.0;

void video_option_vbeline_clock_multiplier_set(double factor) {
	video_option_vbeline_clock_multiplier = factor;
}

double video_option_vbeline_clock_multiplier_get(void) {
	return video_option_vbeline_clock_multiplier;
}

static double video_option_vbeline_8bit_clock_multiplier = 1.0;

void video_option_vbeline_8bit_clock_multiplier_set(double factor) {
	video_option_vbeline_8bit_clock_multiplier = factor;
}

double video_option_vbeline_8bit_clock_multiplier_get(void) {
	return video_option_vbeline_8bit_clock_multiplier;
}

static double video_option_vbeline_16bit_clock_multiplier = 1.0;

void video_option_vbeline_16bit_clock_multiplier_set(double factor) {
	video_option_vbeline_16bit_clock_multiplier = factor;
}

double video_option_vbeline_16bit_clock_multiplier_get(void) {
	return video_option_vbeline_16bit_clock_multiplier;
}

/***************************************************************************/
/* CRCT */

void video_crtc_realize_h(const video_crtc* crtc, unsigned *de, unsigned* bs, unsigned* rs, unsigned* re, unsigned* be, unsigned* tt, unsigned x_div) {
	/* output horizontal */
	*tt = tweak_div_low(crtc->ht,x_div);
	*de = tweak_div_high(crtc->hde,x_div);
	*rs = tweak_div_high(crtc->hrs,x_div);
	*re = tweak_div_low(crtc->hre,x_div);
	if (video_option_crtc_blank_get() == VIDEO_OPTION_CRTC_BLANK_AGGRESSIVE) {
		*bs = *de + 1;
		if (*bs > *rs)
			*bs = *rs;
		*be = *re + 1;
	} else {
		*bs = *rs;
		*be = *re + 1;
	}
}

void video_crtc_realize_v(const video_crtc* crtc, unsigned *de, unsigned* bs, unsigned* rs, unsigned* re, unsigned* be, unsigned* tt, unsigned y_mul) {
	*tt = crtc->vt * y_mul;
	*de = crtc->vde * y_mul;
	*rs = crtc->vrs * y_mul;
	*re = crtc->vre * y_mul;
	if (video_option_crtc_blank_get() == VIDEO_OPTION_CRTC_BLANK_AGGRESSIVE) {
		*bs = *de + 1;
		if (*bs > *rs)
			*bs = *rs;
		*be = *re + 1;
	} else {
		*bs = *rs;
		*be = *re + 1;
	}
}

/* Compute crtc value from vga tweak information
	in:
		info tweak information
		mode video mode
		vert_freq input vertical frequency
	out:
		crtc computed value
	note:
		suppose NOT interlaced mode (VGA regs don't support this feature)
*/
video_error video_crtc_import(video_crtc* crtc, struct tweak_crtc* info, video_mode* mode, double vert_freq) {
	unsigned y_mul = 0;
	unsigned x_div = 0;

	switch (video_mode_driver(mode)) {
		case VIDEO_FLAGS_DRIVER_VGA :
		case VIDEO_FLAGS_DRIVER_VGALINE :
		case VIDEO_FLAGS_DRIVER_VGATWEAK :
			x_div = info->pixels_per_clock;
		break;
		case VIDEO_FLAGS_DRIVER_VBE :
		case VIDEO_FLAGS_DRIVER_VBELINE :
			if (info->hde == mode->size_x)
				x_div = 1;
			else if (info->hde*2 == mode->size_x)
				x_div = 2;
			else if (info->hde*4 == mode->size_x)
				x_div = 4;
			else if (info->hde*8 == mode->size_x)
				x_div = 8;
			else if (info->hde*16 == mode->size_x)
				x_div = 16;
		break;
	}

	if (!x_div) {
		video_error_description_set("Video mode don't support VGA CRTC regs grabbing");
		return -1;
	}

	crtc->hde = info->hde * x_div;
	crtc->hrs = info->hrs * x_div;
	crtc->hre = info->hre * x_div;
	crtc->ht = info->ht * x_div;

	switch (video_mode_driver(mode)) {
		case VIDEO_FLAGS_DRIVER_VGA :
		case VIDEO_FLAGS_DRIVER_VGALINE :
		case VIDEO_FLAGS_DRIVER_VGATWEAK :
			if (info->is_graphics_mode) {
				y_mul = info->scanlines_per_line;
			} else {
				y_mul = 1;
			}
		break;
		case VIDEO_FLAGS_DRIVER_VBE :
		case VIDEO_FLAGS_DRIVER_VBELINE :
			if (info->vde == mode->size_y)
				y_mul = 1;
			else if (info->vde == 2*mode->size_y)
				y_mul = 2;
			else if (info->vde == 4*mode->size_y)
				y_mul = 4;
		break;
	}

	if (!y_mul) {
		video_error_description_set("Video mode don't support VGA CRTC regs grabbing");
		return -1;
	}

	crtc->doublescan = y_mul >= 2;

	crtc->vde = info->vde / y_mul;
	crtc->vrs = info->vrs / y_mul;
	crtc->vre = info->vre / y_mul;
	crtc->vt = info->vt / y_mul;

	/* pixelclock is computed, not imported */
	crtc->pixelclock = crtc->vt * y_mul * crtc->ht * vert_freq;

	crtc->hsync = info->hsync;
	crtc->vsync = info->vsync;

	crtc->interlaced = 0;
	return 0;
}

/***************************************************************************/

void video_mode_done(void) {
	assert( video_mode_active() );

	switch (video_state.mode_info.flags & VIDEO_FLAGS_DRIVER_MASK) {
		case VIDEO_FLAGS_DRIVER_VGATWEAK :
			vgatweak_mode_done();
			break;
		case VIDEO_FLAGS_DRIVER_VGALINE :
			vgaline_mode_done();
			break;
		case VIDEO_FLAGS_DRIVER_VGA :
			vga_mode_done();
			break;
		case VIDEO_FLAGS_DRIVER_VBE :
			vbe_mode_done();
			break;
		case VIDEO_FLAGS_DRIVER_VBELINE :
			vbeline_mode_done();
			break;
		default:
			assert(0);
	}

	video_write_line = 0;
	video_write_selector = 0;

	/* disable */
	video_state.mode_active = 0;
}

static void video_state_rgb_set(void);
static void video_state_rgb_clear(void);

/* Set one video mode
 * return:
 *   ==0 if successfull
 */
video_error video_mode_set(video_mode* mode) {
	if (video_mode_active())
		video_mode_done();

	video_log("video: mode_set %s %dx%d %d %.1f kHz/%.1f Hz\n",video_mode_name(mode),video_mode_size_x(mode),video_mode_size_y(mode),video_mode_bits_per_pixel(mode),(double)video_mode_horz_freq(mode) / 1E3,(double)video_mode_vert_freq(mode));

	switch (mode->flags & VIDEO_FLAGS_DRIVER_MASK) {
		case VIDEO_FLAGS_DRIVER_VGATWEAK :
			if (vgatweak_mode_set(&mode->driver.vgatweak)!=0) {
				return -1;
			}
			video_state.mode_active = 1;
			video_state.mode_info = *mode;
			video_write_line = vga_write_line;
			video_write_selector = vga_write_selector;
		break;
		case VIDEO_FLAGS_DRIVER_VGALINE :
			if (vgaline_mode_set(&mode->driver.vgaline)!=0) {
				return -1;
			}

			video_state.mode_active = 1;
			video_state.mode_info = *mode;
			video_write_line = vga_write_line;
			video_write_selector = vga_write_selector;
		break;
		case VIDEO_FLAGS_DRIVER_VGA :
			if (vga_mode_set(mode->driver.vga.mode,0,0)!=0) {
				return -1;
			}
			video_state.mode_active = 1;
			video_state.mode_info = *mode;
			video_write_line = vga_write_line;
			video_write_selector = vga_write_selector;
		break;
		case VIDEO_FLAGS_DRIVER_VBE :
			if (vbe_mode_set(mode->driver.vbe.mode,0)!=0) {
				return -1;
			}
			video_state.mode_active = 1;
			video_state.mode_info = *mode;
			video_state.mode_info.virtual_x = vbe_state.virtual_x;
			video_state.mode_info.virtual_y = vbe_state.virtual_y;
			video_write_line = vbe_write_line;
			video_write_selector = vbe_write_selector;
		break;
		case VIDEO_FLAGS_DRIVER_VBELINE : {
			if (vbeline_mode_set(&mode->driver.vbeline)!=0) {
				return -1;
			}
			video_state.mode_active = 1;
			video_state.mode_info = *mode;
			video_state.mode_info.virtual_x = vbeline_state.virtual_x;
			video_state.mode_info.virtual_y = vbeline_state.virtual_y;
			video_write_line = vbeline_write_line;
			video_write_selector = vbeline_write_selector;
		} break;
	}

	/* rgb */
	if (video_index() == VIDEO_FLAGS_INDEX_RGB) {
		switch (mode->flags & VIDEO_FLAGS_DRIVER_MASK) {
			case VIDEO_FLAGS_DRIVER_VBE :
			case VIDEO_FLAGS_DRIVER_VBELINE :
				video_state.rgb_red_shift = vbe_state.rgb_red_shift;
				video_state.rgb_green_shift = vbe_state.rgb_green_shift;
				video_state.rgb_blue_shift = vbe_state.rgb_blue_shift;
				video_state.rgb_red_mask = vbe_state.rgb_red_mask;
				video_state.rgb_green_mask = vbe_state.rgb_green_mask;
				video_state.rgb_blue_mask = vbe_state.rgb_blue_mask;
			break;
		}
		video_state_rgb_set();
	} else {
		video_state_rgb_clear();
	}

	{
		double freq = video_measure_vert_freq();
		video_log("video: measured vert clock: %.2f Hz\n",freq);
	}

	return 0;
}

/* Grab the current video mode */
int video_mode_grab(video_mode* mode) {
	if (vbe_mode_grab(&mode->driver.vbe)==0 && vbe_mode_import(mode)==0)
		return 0;
	if (vgatweak_mode_grab(&mode->driver.vgatweak)==0 && vgatweak_mode_import(mode)==0)
		return 0;
	if (vgaline_mode_grab(&mode->driver.vgaline)==0 && vgaline_mode_import(mode)==0)
		return 0;
	return -1;
}

/* Import the specified video mode */
video_error video_mode_import(video_mode* vm) {
	switch (video_mode_driver(vm)) {
		case VIDEO_FLAGS_DRIVER_VGA: return vga_mode_import(vm);
		case VIDEO_FLAGS_DRIVER_VGALINE: return vgaline_mode_import(vm);
		case VIDEO_FLAGS_DRIVER_VGATWEAK: return vgatweak_mode_import(vm);
		case VIDEO_FLAGS_DRIVER_VBE: return vbe_mode_import(vm);
		case VIDEO_FLAGS_DRIVER_VBELINE: return vbeline_mode_import(vm);
	}
	return -1;
}

/* Import the specified video mode and insert in the container */
video_error video_mode_import_and_insert(video_mode_container* vmc, video_mode* vm) {
	switch (video_mode_driver(vm)) {
		case VIDEO_FLAGS_DRIVER_VGA: return vga_mode_import_and_insert(vmc,vm);
		case VIDEO_FLAGS_DRIVER_VGALINE: return vgaline_mode_import_and_insert(vmc,vm);
		case VIDEO_FLAGS_DRIVER_VGATWEAK: return vgatweak_mode_import_and_insert(vmc,vm);
		case VIDEO_FLAGS_DRIVER_VBE: return vbe_mode_import_and_insert(vmc,vm);
		case VIDEO_FLAGS_DRIVER_VBELINE: return vbeline_mode_import_and_insert(vmc,vm);
	}
	return -1;
}

/* Set the startup mode */
void video_mode_reset(void) {
	if (video_mode_active())
		video_mode_done();

	if (video_state.old_mode_grabbed) {
		if (video_mode_set(&video_state.old_mode_info) == 0) {
			video_mode_done();
			return;
		}
	}

	vga_mode_set(0x3,0,0);
	vga_mode_done();
}

int video_mode_vert_freq_adjust_avaliable(video_mode* mode) {
	switch (video_mode_driver(mode)) {
		default:
			assert(0);
		case VIDEO_FLAGS_DRIVER_VGA :
		case VIDEO_FLAGS_DRIVER_VGALINE :
		case VIDEO_FLAGS_DRIVER_VGATWEAK :
		case VIDEO_FLAGS_DRIVER_VBE :
			return 0;
		case VIDEO_FLAGS_DRIVER_VBELINE :
			return 1;
	}
}

void video_mode_vert_freq_adjust(video_mode* mode, float vsync) {
	switch (video_mode_driver(mode)) {
		default:
			assert(0);
		case VIDEO_FLAGS_DRIVER_VGA :
		case VIDEO_FLAGS_DRIVER_VGALINE :
		case VIDEO_FLAGS_DRIVER_VGATWEAK :
		case VIDEO_FLAGS_DRIVER_VBE :
			break;
		case VIDEO_FLAGS_DRIVER_VBELINE : {
				mode->driver.vbeline.crtc.pixelclock = vsync * mode->driver.vbeline.crtc.ht * mode->driver.vbeline.crtc.vt;
				if (mode->driver.vbeline.crtc.doublescan)
					mode->driver.vbeline.crtc.pixelclock *= 2;
				if (mode->driver.vbeline.crtc.interlaced)
					mode->driver.vbeline.crtc.pixelclock /= 2;
				mode->vert_freq = vsync;
				mode->horz_freq = (double)mode->driver.vbeline.crtc.pixelclock / mode->driver.vbeline.crtc.ht;
			}
			break;
	}
}

/* Wait for a vertical retrace */
void video_wait_vsync(void) {
	switch (video_driver()) {
		default:
			assert(0);
		case VIDEO_FLAGS_DRIVER_VGA :
		case VIDEO_FLAGS_DRIVER_VGALINE :
		case VIDEO_FLAGS_DRIVER_VGATWEAK :
			vga_wait_vsync();
			break;
		case VIDEO_FLAGS_DRIVER_VBE :
			vbe_wait_vsync();
			break;
		case VIDEO_FLAGS_DRIVER_VBELINE :
			vbeline_wait_vsync();
			break;
	}
}

/* Measure the frame rate */
float video_measure_vert_freq(void) {
	double vert_freq = 0;
	unsigned retry = 4;

	uclock(); /* reset clock */

	do {
		clock_t start, stop;
		video_wait_vsync();
		start = uclock();
		video_wait_vsync();
		stop = uclock();
		if (start != stop) {
			vert_freq = (double)UCLOCKS_PER_SEC / (stop - start);
		}
		--retry;
	} while (retry && (vert_freq < 10 || vert_freq > 300));

	if (!retry)
		return 0;

	return vert_freq;
}

/* Set the visual offset, return immediatially */
video_error video_display_set_async(unsigned offset, int waitvsync) {
	assert( video_mode_active() );

	switch (video_driver()) {
		default:
			assert(0);
		case VIDEO_FLAGS_DRIVER_VGA :
		case VIDEO_FLAGS_DRIVER_VGALINE :
		case VIDEO_FLAGS_DRIVER_VGATWEAK :
			vga_scroll(offset,waitvsync);
			break;
		case VIDEO_FLAGS_DRIVER_VBE :
		case VIDEO_FLAGS_DRIVER_VBELINE :
			if (vbe_scroll(offset,waitvsync)!=0) {
				return -1;
			}
			break;
	}
	return 0;
}

/* Set the visual offset at next vsync, return immeditially */
video_error video_display_set_sync(unsigned offset) {
	assert( video_mode_active() );

	switch (video_driver()) {
		default:
			video_error_description_set("Syncronous scroll unsupported with this video mode");
			return -1;
		case VIDEO_FLAGS_DRIVER_VBE :
		case VIDEO_FLAGS_DRIVER_VBELINE :
			if (vbe_request_scroll(offset)!=0) {
				return -1;
			}
			break;
	}
	return 0;
}

unsigned video_bytes_per_page(void) {
	unsigned bytes_per_page = video_size_y() * video_bytes_per_scanline();
	switch (video_driver()) {
		default:
			assert(0);
		case VIDEO_FLAGS_DRIVER_VGA :
		case VIDEO_FLAGS_DRIVER_VGALINE :
		case VIDEO_FLAGS_DRIVER_VGATWEAK :
			return vga_adjust_bytes_per_page(bytes_per_page);
		case VIDEO_FLAGS_DRIVER_VBE :
		case VIDEO_FLAGS_DRIVER_VBELINE :
			return vbe_adjust_bytes_per_page(bytes_per_page);
	}
}

static __inline__ void video_chained_put_pixel(unsigned x, unsigned y, unsigned color) {
	switch (video_bytes_per_pixel()) {
		case 1 :
			_farpokeb( video_write_selector(), video_write_line(y) + x, color);
		break;
		case 2 :
			_farpokew( video_write_selector(), video_write_line(y) + 2*x, color);
		break;
		case 4 :
			_farpokel( video_write_selector(), video_write_line(y) + 4*x, color);
		break;
	}
}

static __inline__ void video_unchained_put_pixel(unsigned x, unsigned y, unsigned color) {
	vga_unchained_plane_set(x % 4);
	_farpokeb( vga_write_selector(), vga_write_line(y) + x / 4, color);
}

void video_put_char(unsigned x, unsigned y, char c, unsigned color) {
	assert( video_is_text() );
	if (vga_mode_active())
		_farpokew( vga_write_selector(), vga_write_line(y) + 2*x, (color << 8) | (unsigned char)c);
	else
		_farpokew( vbe_write_selector(), vbe_write_line(y) + 2*x, (color << 8) | (unsigned char)c);
}

static __inline__ void video_put_pixel_noclip(unsigned x, unsigned y, unsigned color) {
	assert( video_is_graphics() );
	if (video_unchained_active())
		video_unchained_put_pixel(x,y,color);
	else
		video_chained_put_pixel(x,y,color);
}

void video_put_pixel(unsigned x, unsigned y, unsigned color) {
	if (x < video_virtual_x() && y < video_virtual_y())
		video_put_pixel_noclip(x,y,color);
}

/****************************************************************************/
/* color */

/* Compute the distance of two color */
unsigned video_color_dist(const video_color* A, const video_color* B) {
	int r,g,b;
	r = (int)A->red - B->red;
	g = (int)A->green - B->green;
	b = (int)A->blue - B->blue;
	return r*r + g*g + b*b;
}

/* Compute the size in bit of the mask */
static unsigned video_rgb_len(unsigned mask) {
	unsigned len = 0;
	while ((mask & 1) == 0)
		mask >>= 1;
	while (mask) {
		++len;
		mask >>= 1;
	}
	return len;
}

/* Update the video_state.rgb_* variables */
static void video_state_rgb_set(void) {
	video_state.rgb_red_len = video_rgb_len(video_state.rgb_red_mask);
	video_state.rgb_green_len = video_rgb_len(video_state.rgb_green_mask);
	video_state.rgb_blue_len = video_rgb_len(video_state.rgb_blue_mask);
	video_state.rgb_mask_bit = video_rgb_get(0xFF,0xFF,0xFF);
	video_state.rgb_high_bit = video_rgb_get(0x80,0x80,0x80);
	video_state.rgb_low_bit =  video_rgb_get(
		1 << (8-video_state.rgb_red_len),
		1 << (8-video_state.rgb_green_len),
		1 << (8-video_state.rgb_blue_len));
}

static void video_state_rgb_clear(void) {
	video_state.rgb_red_shift = 0;
	video_state.rgb_green_shift = 0;
	video_state.rgb_blue_shift = 0;
	video_state.rgb_red_mask = 0;
	video_state.rgb_green_mask = 0;
	video_state.rgb_blue_mask = 0;
	video_state.rgb_red_len = 0;
	video_state.rgb_green_len = 0;
	video_state.rgb_blue_len = 0;
	video_state.rgb_mask_bit = 0;
	video_state.rgb_high_bit = 0;
	video_state.rgb_low_bit =  0;
}

static unsigned video_rgb_approx(unsigned value, unsigned len) {
	unsigned fill = len;
	while (fill < 8) {
		value |= value >> fill;
		fill *= 2;
	}
	return value;
}

unsigned video_red_get_approx(unsigned rgb) {
	return video_rgb_approx(video_red_get(rgb), video_state.rgb_red_len);
}

unsigned video_green_get_approx(unsigned rgb) {
	return video_rgb_approx(video_green_get(rgb), video_state.rgb_green_len);
}

unsigned video_blue_get_approx(unsigned rgb) {
	return video_rgb_approx(video_blue_get(rgb), video_state.rgb_blue_len);
}

/* Compute shift and mask value */
static __inline__ void video_rgb_data(int* shift, unsigned* mask, unsigned posbit, unsigned masksize) {
	*mask = ((1 << masksize) - 1) << posbit;
	*shift = posbit + masksize - 8;
}

int video_index_packed_to_rgb_avaliable(void) {
	return video_index() == VIDEO_FLAGS_INDEX_PACKED;
}

/* Convert one video mode from PACKED to RGB */
void video_index_packed_to_rgb(int waitvsync) {
	unsigned bit_r;
	unsigned bit_g;
	unsigned bit_b;

	video_color* palette;
	unsigned i;
	unsigned colors = 1 << video_bits_per_pixel();

	assert( video_index_packed_to_rgb_avaliable() );

	bit_b = video_bits_per_pixel() / 3;
	bit_r = (video_bits_per_pixel() - bit_b) / 2;
	bit_g = video_bits_per_pixel() - bit_b - bit_r;

	video_rgb_data(&video_state.rgb_red_shift,&video_state.rgb_red_mask,0,bit_r);
	video_rgb_data(&video_state.rgb_green_shift,&video_state.rgb_green_mask,bit_r,bit_g);
	video_rgb_data(&video_state.rgb_blue_shift,&video_state.rgb_blue_mask,bit_r+bit_g,bit_b);

	video_state_rgb_set();

	palette = malloc(colors * sizeof(video_color));
	assert( palette );

	for(i=0;i<colors;++i) {
		palette[i].red = video_red_get_approx(i);
		palette[i].green = video_green_get_approx(i);
		palette[i].blue = video_blue_get_approx(i);
	}

	video_palette_set(palette,0,colors,waitvsync);

	free(palette);

	/* set RGB */
	video_state.mode_info.flags &= ~VIDEO_FLAGS_INDEX_MASK;
	video_state.mode_info.flags |= VIDEO_FLAGS_INDEX_RGB;
}

int video_index_rgb_to_packed_avaliable(void) {
	return video_index() == VIDEO_FLAGS_INDEX_RGB && video_bytes_per_pixel() == 1;
}

/* Convert one video mode from RGB to PACKED */
void video_index_rgb_to_packed(void) {
	assert( video_index_rgb_to_packed_avaliable() );

	video_state_rgb_clear();

	/* set PACKED */
	video_state.mode_info.flags &= ~VIDEO_FLAGS_INDEX_MASK;
	video_state.mode_info.flags |= VIDEO_FLAGS_INDEX_PACKED;
}

/* last active palette */
static video_color video_palette[256];

video_color* video_palette_get(void) {
	return video_palette;
}

video_error video_palette_set(video_color* palette, unsigned start, unsigned count, int waitvsync) {
	assert( (video_flags() & VIDEO_FLAGS_INDEX_MASK) == VIDEO_FLAGS_INDEX_PACKED );
	assert( count <= 256 );

	/* update the static palette */
	memcpy(video_palette + start, palette, count * sizeof(video_color));

	switch (video_driver()) {
		default:
			assert(0);
		case VIDEO_FLAGS_DRIVER_VGA :
		case VIDEO_FLAGS_DRIVER_VGALINE :
		case VIDEO_FLAGS_DRIVER_VGATWEAK :
			vga_palette8_set(palette,start,count,waitvsync);
			break;
		case VIDEO_FLAGS_DRIVER_VBE :
			if (vbe_palette8_set(palette,start,count,waitvsync)!=0)
				return -1;
			break;
		case VIDEO_FLAGS_DRIVER_VBELINE :
			if (vbeline_palette8_set(palette,start,count,waitvsync)!=0)
				return -1;
			break;
	}

	return 0;
}

/****************************************************************************/
/* error */

void video_error_description_set(const char* error,...) {
	va_list arg;
	va_start(arg,error);
	vsprintf(video_state.error,error,arg);
	video_log("video: set_error_description \"");
	video_log_va(error,arg);
	video_log("\"\n");
	va_end(arg);
}

void video_error_description_nolog_set(const char* error,...) {
	va_list arg;
	va_start(arg,error);
	vsprintf(video_state.error,error,arg);
	va_end(arg);
}

/****************************************************************************/
/* log */

int video_print_modeline_mc(char* dest, unsigned mode, const video_crtc* crtc) {
	const char* flag1 = crtc->hsync ? " -hsync" : " +hsync";
	const char* flag2 = crtc->vsync ? " -vsync" : " +vsync";
	const char* flag3 = crtc->doublescan ? " doublescan" : "";
	const char* flag4 = crtc->interlaced ? " interlaced" : "";
	return sprintf(dest,"%x %.4f %d %d %d %d %d %d %d %d%s%s%s%s",
		mode, (double)crtc->pixelclock / 1E6,
		crtc->hde, crtc->hrs, crtc->hre, crtc->ht,
		crtc->vde, crtc->vrs, crtc->vre, crtc->vt,
		flag1, flag2, flag3, flag4
	);
}

void video_log_modeline_mb(const char *text, unsigned mode, unsigned pixel_clock, unsigned hde, unsigned hbs, unsigned hrs, unsigned hre, unsigned hbe, unsigned ht, unsigned vde, unsigned vbs, unsigned vrs, unsigned vre, unsigned vbe, unsigned vt, int hsync_pol, int vsync_pol, int doublescan, int interlace) {
	const char* flag1 = hsync_pol ? " -hsync" : " +hsync";
	const char* flag2 = vsync_pol ? " -vsync" : " +vsync";
	const char* flag3 = doublescan ? " doublescan" : "";
	const char* flag4 = interlace ? " interlaced" : "";
	video_log("%s(%x %.4f %d %d %d %d %d %d %d %d %d %d %d %d%s%s%s%s)\n",
		text, mode, (double)pixel_clock / 1E6,
		hde, hbs, hrs, hre, hbe, ht,
		vde, vbs, vrs, vre, vbe, vt,
		flag1, flag2, flag3, flag4
	);
}

void video_log_modeline_m(const char *text, unsigned mode, unsigned pixel_clock, unsigned hde, unsigned hrs, unsigned hre, unsigned ht, unsigned vde, unsigned vrs, unsigned vre, unsigned vt, int hsync_pol, int vsync_pol, int doublescan, int interlace) {
	const char* flag1 = hsync_pol ? " -hsync" : " +hsync";
	const char* flag2 = vsync_pol ? " -vsync" : " +vsync";
	const char* flag3 = doublescan ? " doublescan" : "";
	const char* flag4 = interlace ? " interlaced" : "";
	video_log("%s(%x %.4f %d %d %d %d %d %d %d %d%s%s%s%s)\n",
		text, mode, (double)pixel_clock / 1E6,
		hde, hrs, hre, ht,
		vde, vrs, vre, vt,
		flag1, flag2, flag3, flag4
	);
}

void video_log_modeline_b(const char *text, unsigned pixel_clock, unsigned hde, unsigned hbs, unsigned hrs, unsigned hre, unsigned hbe, unsigned ht, unsigned vde, unsigned vbs, unsigned vrs, unsigned vre, unsigned vbe, unsigned vt, int hsync_pol, int vsync_pol, int doublescan, int interlace) {
	const char* flag1 = hsync_pol ? " -hsync" : " +hsync";
	const char* flag2 = vsync_pol ? " -vsync" : " +vsync";
	const char* flag3 = doublescan ? " doublescan" : "";
	const char* flag4 = interlace ? " interlaced" : "";
	video_log("%s(%.4f %d %d %d %d %d %d %d %d %d %d %d %d%s%s%s%s)\n",
		text, (double)pixel_clock / 1E6,
		hde, hbs, hrs, hre, hbe, ht,
		vde, vbs, vrs, vre, vbe, vt,
		flag1, flag2, flag3, flag4
	);
}

void video_log(const char* text,...) {
	va_list arg;
	va_start(arg,text);
	video_log_va(text,arg);
	va_end(arg);
}
