/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2000 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 */

#include "vbe.h"

#include <stdlib.h>
#include <stdio.h>

int vbe_mode_grab(vbe_video_mode* mode) {
	if (!vbe_active())
		return -1;

	if (vbe_mode_get(&mode->mode)!=0) {
		return -1;
	}

	if (mode->mode < 0x100)
		return -1;

	return 0;
}

/* Import information for one video mode
	vm->vbe video mode
	Return 0 if successful
*/
int vbe_mode_import(video_mode* mode) {
	vbe_ModeInfoBlock info;

	if (vbe_mode_info_get(&info, mode->driver.vbe.mode)!=0) {
		video_error_description_set("VBE report that mode %d is unsupported", mode->driver.vbe.mode);
		return -1;
	}
	if ((info.ModeAttributes & vbeMdAvailable) == 0) {
		video_error_description_set("VBE report that mode %d is not avaliable", mode->driver.vbe.mode);
		return -1;
	}

	if ((info.ModeAttributes & vbeMdGraphMode) == 0) {
		if (info.MemoryModel!=vbeMemTXT) {
			video_error_description_nolog_set("Unsupported memory model");
			return -1;
		}
		// TODO
		video_error_description_nolog_set("Text modes are unsupported");
		return -1;
	} else {
		if ((mode->driver.vbe.mode & vbeLinearBuffer) != 0) {
			if (vbe_state.info.VESAVersion < 0x200) {
				video_error_description_nolog_set("Linear frame buffer not available on VBE version prior 2.0");
				return -1;
			}
			if ((info.ModeAttributes & vbeMdLinear) == 0) {
				video_error_description_nolog_set("Linear frame buffer not available in this mode");
				return -1;
			}
		} else {
			if ((info.ModeAttributes & vbeMdNonBanked) != 0) {
				video_error_description_nolog_set("Banked frame buffer not available in this mode");
				return -1;
			}
		}
		/* 8,15,16 bits per pixel */
		if (info.BitsPerPixel!=8 && info.BitsPerPixel!=15 && info.BitsPerPixel!=16) {
			video_error_description_nolog_set("Unsupported bits per pixel size");
			return -1;
		}
		/* Packed or RGB memory model */
		if (info.MemoryModel!=vbeMemPK && info.MemoryModel!=vbeMemRGB) {
			video_error_description_nolog_set("Unsupported memory model");
			return -1;
		}
		/* Non planar mode */
		if (info.NumberOfPlanes!=1) {
			video_error_description_nolog_set("Unsupported number of planes");
			return -1;
		}
	}
	mode->flags = VIDEO_FLAGS_DRIVER_VBE |
		VIDEO_FLAGS_ASYNC_SETPAGE |
		(mode->flags & VIDEO_FLAGS_USER_MASK);
	if ((info.ModeAttributes & vbeMdTripleBuffer) != 0)
		mode->flags |= VIDEO_FLAGS_SYNC_SETPAGE;
	if (info.MemoryModel==vbeMemTXT)
		mode->flags |= VIDEO_FLAGS_INDEX_TEXT | VIDEO_FLAGS_TYPE_TEXT;
	if (info.MemoryModel==vbeMemPK)
		mode->flags |= VIDEO_FLAGS_INDEX_PACKED | VIDEO_FLAGS_TYPE_GRAPHICS;
	if (info.MemoryModel==vbeMemRGB)
		mode->flags |= VIDEO_FLAGS_INDEX_RGB | VIDEO_FLAGS_TYPE_GRAPHICS;
	if (mode->driver.vbe.mode & vbeLinearBuffer)
		mode->flags |= VIDEO_FLAGS_LINEAR;

	mode->size_x = info.XResolution;
	mode->size_y = info.YResolution;
	if (info.MemoryModel==vbeMemTXT) {
		mode->size_x *= info.XCharSize;
		mode->size_y *= info.YCharSize;
	}
	mode->virtual_x = 0;
	mode->virtual_y = 0;
	mode->vert_freq = 0;
	mode->horz_freq = 0;
	mode->bits = info.BitsPerPixel;
	mode->aspect_x = 4 * mode->size_y;
	mode->aspect_y = 3 * mode->size_x;
	return 0;
}

video_error vbe_mode_import_and_insert(video_mode_container* vmc, video_mode* mode) {
	video_mode mode8 = *mode;
	if (video_mode_import(&mode8)==0) {
		video_mode_container_insert(vmc,&mode8);
	}
	return 0;
}

/* Add all vbe modes */
static void video_mode_container_insert_vbe_all(video_mode_container* vmc) {
	if (vbe_active() && vbe_state.info.VESAVersion >= 0x0102) {
		vbe_mode_iterator vmi;
		vbe_mode_iterator_begin(&vmi);
		while (!vbe_mode_iterator_end(&vmi)) {
			video_mode v;
			memset(&v,0,sizeof(v));
			v.driver.vbe.mode = vbe_mode_iterator_get(&vmi);

			if (vbe_mode_import(&v)==0) {
				if (video_mode_is_text(&v))
					sprintf(v.name,"bios_vbetext_%dx%d",(unsigned)v.size_x,(unsigned)v.size_y);
				else
					sprintf(v.name,"bios_vbebanked_%dx%dx%d",(unsigned)v.size_x,(unsigned)v.size_y,(unsigned)v.bits);
				video_mode_container_insert(vmc,&v);
			}

			if (vbe_state.info.VESAVersion >= 0x200) {
				v.driver.vbe.mode |= vbeLinearBuffer;
				if (vbe_mode_import(&v)==0) {
					if (video_mode_is_graphics(&v)) {
						sprintf(v.name,"bios_vbelinear_%dx%dx%d",(unsigned)v.size_x,(unsigned)v.size_y,(unsigned)v.bits);
						video_mode_container_insert(vmc,&v);
					}
				}
			}

			vbe_mode_iterator_next(&vmi);
		}
	}
}

/* Add vbe modes */
void video_mode_container_insert_vbe(video_mode_container* vmc) {
	video_mode_container all;
	video_mode_container_iterator i;

	if (video_option_default_mode_get() == VIDEO_OPTION_DEFAULT_MODE_ALL) {
		video_mode_container_insert_vbe_all(vmc);
		return;
	}

	video_mode_container_init(&all);
	video_mode_container_insert_vbe_all(&all);

	video_mode_container_iterator_begin(&i,&all);
	while (!video_mode_container_iterator_end(&i)) {
		video_mode* ip = video_mode_container_iterator_get(&i);

		video_mode_container_iterator j;
		video_mode_container_iterator_begin(&j,&all);
		while (!video_mode_container_iterator_end(&j)) {
			video_mode* jp = video_mode_container_iterator_get(&j);
			if (video_mode_size_x(ip) == video_mode_size_x(jp) && video_mode_size_y(ip) == video_mode_size_y(jp) && video_mode_bytes_per_pixel(ip) == video_mode_bytes_per_pixel(jp)) {
				if (!(video_mode_flags(ip) & VIDEO_FLAGS_LINEAR) && (video_mode_flags(jp) & VIDEO_FLAGS_LINEAR))
					break;
				if (video_mode_bits_per_pixel(ip) < video_mode_bits_per_pixel(jp))
					break;
			}

			video_mode_container_iterator_next(&j);
		}

		if (video_mode_container_iterator_end(&j)) {
			video_mode_container_insert(vmc,ip);
		}

		video_mode_container_iterator_next(&i);
	}

	video_mode_container_done(&all);
}

void video_mode_vbe_print(char* s, const video_mode* vm) {
	sprintf(s,"%s vbe(%x)",vm->name,vm->driver.vbe.mode);
}

int vbe_palette8_set(video_color* palette, unsigned start, unsigned count, int waitvsync) {
	video_color vbe_pal[256];
	unsigned shift = 8 - vbe_state.palette_width;

	if (shift) {
		unsigned i;
		for(i=0;i<count;++i) {
			vbe_pal[i].red = palette[i].red >> shift;
			vbe_pal[i].green = palette[i].green >> shift;
			vbe_pal[i].blue = palette[i].blue >> shift;
			vbe_pal[i].alpha = 0;
		}
	} else {
		unsigned i;
		for(i=0;i<count;++i) {
			vbe_pal[i].red = palette[i].red;
			vbe_pal[i].green = palette[i].green;
				vbe_pal[i].blue = palette[i].blue;
				vbe_pal[i].alpha = 0;
		}
	}

	return vbe_palette_set(vbe_pal,start,count,waitvsync);
}

